"""
    Theme data
"""

from builtins import object
from rest_framework.reverse import reverse
from django.utils.translation import ugettext


class Theme(object):
    def __init__(self, pk, name, text, background, border, selected):
        self.id = self.pk = pk
        self.name = name
        self.text = text
        self.background = background
        self.border = border
        self.selected = selected


THEME_DATA = {
    "grass": Theme(
        name="grass",
        text=ugettext("Grass"),
        selected="#bce369",
        background="#a0c53a",
        border="#70961f",
        pk=1,
    ),
    "snow": Theme(
        name="snow",
        text=ugettext("Snow"),
        selected="#b3deff",
        background="#eef7ff",
        border="#83c9fe",
        pk=2,
    ),
    "farm": Theme(
        name="farm",
        text=ugettext("Farm"),
        selected="#bce369",
        background="#a0c53a",
        border="#70961f",
        pk=3,
    ),
    "city": Theme(
        name="city",
        text=ugettext("City"),
        selected="#C1C1C1",
        background="#969696",
        border="#686868",
        pk=4,
    ),
}


def get_theme(name):
    """Helper method to get a theme."""
    return THEME_DATA[name]


def get_all_themes():
    return list(THEME_DATA.values())


def get_theme_by_pk(pk):
    for theme in list(THEME_DATA.values()):
        if theme.pk == int(pk):
            return theme
    raise KeyError


def get_themes_url(pk, request):
    return reverse("theme-detail", args={pk}, request=request)



import csv

from django.http import HttpResponse

Headers = ["Class", "Student", "Completed Levels", "Total Time", "Total Score"]
SharedHeaders = ["Class", "Student"]


def scoreboard_csv(student_data, requested_sorted_levels, improvement_data, shared_level_headers, shared_student_data):
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="scoreboard.csv"'

    improvement_dict = {}
    for student in improvement_data:
        idx = f"{student.class_field.name}_{student.name}"
        improvement_dict[idx] = student.areas

    headers = Headers + requested_sorted_levels + ["Areas for improvement"]
    shared_headers = SharedHeaders + shared_level_headers

    rows = []
    for student in student_data:
        result = [student.class_field.name, student.name, student.completed, student.total_time, student.total_score]

        scores = []
        for level_id, level_score in student.level_scores.items():
            scores.append(level_score["score"])

        row = result + scores

        # add the areas to improve, if there's any
        idx = f"{student.class_field.name}_{student.name}"
        row.append(improvement_dict.get(idx))

        rows.append(row)

    shared_rows = []
    for student in shared_student_data:
        result = [student.class_field.name, student.name]

        scores = []
        for level_id, level_score in student.level_scores.items():
            scores.append(level_score["score"])

        row = result + scores

        shared_rows.append(row)

    writer = csv.writer(response)
    writer.writerow(headers)
    writer.writerows(rows)

    writer.writerow("")

    writer.writerow(shared_headers)
    writer.writerows(shared_rows)

    return response



import json

from django.apps.registry import Apps
from django.db import migrations

from game.level_management import set_decor_inner, set_blocks_inner


def add_loop_levels(apps: Apps, *args):
    Level = apps.get_model("game", "Level")
    LevelBlock = apps.get_model("game", "LevelBlock")
    Block = apps.get_model("game", "Block")
    LevelDecor = apps.get_model("game", "LevelDecor")
    Episode = apps.get_model("game", "Episode")

    episode_15 = Episode.objects.create(
        pk=15,
        name="For Loops - coming soon",
    )
    episode_14 = Episode.objects.create(
        pk=14,
        next_episode=episode_15,
        name="Selection in a Loop - coming soon",
    )
    episode_13 = Episode.objects.create(
        pk=13,
        next_episode=episode_14,
        name="Indeterminate WHILE Loops - coming soon",
    )
    episode_12 = Episode.objects.create(
        pk=12,
        next_episode=episode_13,
        name="Sequencing and Counted Loops",
    )

    episode_11 = Episode.objects.get(pk=11)
    episode_11.next_episode = episode_12
    episode_11.save()

    # Episode 15, Levels 149 - 153
    level_153 = Level.objects.create(
        name="153",
        episode=episode_15,
    )
    level_152 = Level.objects.create(
        name="152",
        episode=episode_15,
        next_level=level_153,
    )
    level_151 = Level.objects.create(
        name="151",
        episode=episode_15,
        next_level=level_152,
    )
    level_150 = Level.objects.create(
        name="150",
        episode=episode_15,
        next_level=level_151,
    )
    level_149 = Level.objects.create(
        name="149",
        episode=episode_15,
        next_level=level_150,
    )

    # Episode 14, Levels 141 - 148
    level_148 = Level.objects.create(
        name="148",
        episode=episode_14,
        next_level=level_149,
    )
    level_147 = Level.objects.create(
        name="147",
        episode=episode_14,
        next_level=level_148,
    )
    level_146 = Level.objects.create(
        name="146",
        episode=episode_14,
        next_level=level_147,
    )
    level_145 = Level.objects.create(
        name="145",
        episode=episode_14,
        next_level=level_146,
    )
    level_144 = Level.objects.create(
        name="144",
        episode=episode_14,
        next_level=level_145,
    )
    level_143 = Level.objects.create(
        name="143",
        episode=episode_14,
        next_level=level_144,
    )
    level_142 = Level.objects.create(
        name="142",
        episode=episode_14,
        next_level=level_143,
    )
    level_141 = Level.objects.create(
        name="141",
        episode=episode_14,
        next_level=level_142,
    )

    # Episode 13, Levels 123 - 140
    level_140 = Level.objects.create(
        name="140",
        episode=episode_13,
        next_level=level_141,
    )
    level_139 = Level.objects.create(
        name="139",
        episode=episode_13,
        next_level=level_140,
    )
    level_138 = Level.objects.create(
        name="138",
        episode=episode_13,
        next_level=level_139,
    )
    level_137 = Level.objects.create(
        name="137",
        episode=episode_13,
        next_level=level_138,
    )
    level_136 = Level.objects.create(
        name="136",
        episode=episode_13,
        next_level=level_137,
    )
    level_135 = Level.objects.create(
        name="135",
        episode=episode_13,
        next_level=level_136,
    )
    level_134 = Level.objects.create(
        name="134",
        episode=episode_13,
        next_level=level_135,
    )
    level_133 = Level.objects.create(
        name="133",
        episode=episode_13,
        next_level=level_134,
    )
    level_132 = Level.objects.create(
        name="132",
        episode=episode_13,
        next_level=level_133,
    )
    level_131 = Level.objects.create(
        name="131",
        episode=episode_13,
        next_level=level_132,
    )
    level_130 = Level.objects.create(
        name="130",
        episode=episode_13,
        next_level=level_131,
    )
    level_129 = Level.objects.create(
        name="129",
        episode=episode_13,
        next_level=level_130,
    )
    level_128 = Level.objects.create(
        name="128",
        episode=episode_13,
        next_level=level_129,
    )
    level_127 = Level.objects.create(
        name="127",
        episode=episode_13,
        next_level=level_128,
    )
    level_126 = Level.objects.create(
        name="126",
        episode=episode_13,
        next_level=level_127,
    )
    level_125 = Level.objects.create(
        name="125",
        episode=episode_13,
        next_level=level_126,
    )
    level_124 = Level.objects.create(
        name="124",
        episode=episode_13,
        next_level=level_125,
    )
    level_123 = Level.objects.create(
        name="123",
        episode=episode_13,
        next_level=level_124,
    )

    # Episode 12, Levels 110 - 122
    level_122 = Level.objects.create(
        name="122",
        episode=episode_12,
        # next_level=level_123, TODO: connect them when the next levels are enabled.
        path='[{"coordinate":[1,5],"connectedNodes":[1]},{"coordinate":[1,4],"connectedNodes":[0,2]},{"coordinate":[2,4],"connectedNodes":[1,3]},{"coordinate":[2,5],"connectedNodes":[4,2]},{"coordinate":[3,5],"connectedNodes":[3,5]},{"coordinate":[3,4],"connectedNodes":[4,6]},{"coordinate":[4,4],"connectedNodes":[5,7]},{"coordinate":[4,5],"connectedNodes":[8,6]},{"coordinate":[5,5],"connectedNodes":[7,9]},{"coordinate":[5,4],"connectedNodes":[8,10]},{"coordinate":[6,4],"connectedNodes":[9,11]},{"coordinate":[6,5],"connectedNodes":[12,10]},{"coordinate":[7,5],"connectedNodes":[11,13]},{"coordinate":[7,4],"connectedNodes":[12,14]},{"coordinate":[7,3],"connectedNodes":[13,15]},{"coordinate":[7,2],"connectedNodes":[14,16]},{"coordinate":[7,1],"connectedNodes":[17,15]},{"coordinate":[6,1],"connectedNodes":[18,16]},{"coordinate":[6,2],"connectedNodes":[19,17]},{"coordinate":[5,2],"connectedNodes":[18,20]},{"coordinate":[5,1],"connectedNodes":[21,19]},{"coordinate":[4,1],"connectedNodes":[22,20]},{"coordinate":[4,2],"connectedNodes":[23,21]},{"coordinate":[3,2],"connectedNodes":[22,24]},{"coordinate":[3,1],"connectedNodes":[25,23]},{"coordinate":[2,1],"connectedNodes":[26,24]},{"coordinate":[2,2],"connectedNodes":[27,25]},{"coordinate":[1,2],"connectedNodes":[26,28]},{"coordinate":[1,1],"connectedNodes":[27]}]',
        origin='{"coordinate":[1,5],"direction":"S"}',
        destinations="[[1,1]]",
        default=True,
        blocklyEnabled=False,
        theme_name="grass",
        character_name="Van",
    )
    level_121 = Level.objects.create(
        name="121",
        episode=episode_12,
        next_level=level_122,
        path='[{"coordinate":[1,7],"connectedNodes":[1]},{"coordinate":[1,6],"connectedNodes":[0,2,13]},{"coordinate":[2,6],"connectedNodes":[1,25,3]},{"coordinate":[2,5],"connectedNodes":[2,4]},{"coordinate":[3,5],"connectedNodes":[3,5]},{"coordinate":[3,4],"connectedNodes":[4,6]},{"coordinate":[4,4],"connectedNodes":[5,7]},{"coordinate":[4,3],"connectedNodes":[23,6,8]},{"coordinate":[5,3],"connectedNodes":[7,9]},{"coordinate":[5,2],"connectedNodes":[8,10]},{"coordinate":[6,2],"connectedNodes":[9,31,11]},{"coordinate":[6,1],"connectedNodes":[21,10,12]},{"coordinate":[7,1],"connectedNodes":[11,32]},{"coordinate":[1,5],"connectedNodes":[1,14]},{"coordinate":[1,4],"connectedNodes":[13,15]},{"coordinate":[1,3],"connectedNodes":[14,22,16]},{"coordinate":[1,2],"connectedNodes":[15,17]},{"coordinate":[1,1],"connectedNodes":[16,18]},{"coordinate":[2,1],"connectedNodes":[17,19]},{"coordinate":[3,1],"connectedNodes":[18,24,20]},{"coordinate":[4,1],"connectedNodes":[19,21]},{"coordinate":[5,1],"connectedNodes":[20,11]},{"coordinate":[2,3],"connectedNodes":[15,23]},{"coordinate":[3,3],"connectedNodes":[22,7,24]},{"coordinate":[3,2],"connectedNodes":[23,19]},{"coordinate":[3,6],"connectedNodes":[2,26]},{"coordinate":[4,6],"connectedNodes":[25,27]},{"coordinate":[5,6],"connectedNodes":[26,28]},{"coordinate":[6,6],"connectedNodes":[27,29]},{"coordinate":[6,5],"connectedNodes":[28,30]},{"coordinate":[6,4],"connectedNodes":[29,31]},{"coordinate":[6,3],"connectedNodes":[30,10]},{"coordinate":[7,0],"connectedNodes":[12]}]',
        origin='{"coordinate":[1,7],"direction":"S"}',
        destinations="[[7,0]]",
        default=True,
        blocklyEnabled=False,
        theme_name="farm",
        character_name="Van",
    )
    level_120 = Level.objects.create(
        name="120",
        episode=episode_12,
        next_level=level_121,
        path='[{"coordinate":[9,5],"connectedNodes":[1]},{"coordinate":[8,5],"connectedNodes":[2,0]},{"coordinate":[7,5],"connectedNodes":[1,3]},{"coordinate":[7,4],"connectedNodes":[4,2,20]},{"coordinate":[6,4],"connectedNodes":[5,3]},{"coordinate":[5,4],"connectedNodes":[13,4,6]},{"coordinate":[5,3],"connectedNodes":[7,5]},{"coordinate":[4,3],"connectedNodes":[8,6]},{"coordinate":[3,3],"connectedNodes":[7,9]},{"coordinate":[3,2],"connectedNodes":[10,8,24]},{"coordinate":[2,2],"connectedNodes":[11,9]},{"coordinate":[1,2],"connectedNodes":[19,10,12]},{"coordinate":[1,1],"connectedNodes":[25,11]},{"coordinate":[5,5],"connectedNodes":[14,5]},{"coordinate":[4,5],"connectedNodes":[15,13]},{"coordinate":[3,5],"connectedNodes":[16,14]},{"coordinate":[2,5],"connectedNodes":[17,15]},{"coordinate":[1,5],"connectedNodes":[16,18]},{"coordinate":[1,4],"connectedNodes":[17,19]},{"coordinate":[1,3],"connectedNodes":[18,11]},{"coordinate":[7,3],"connectedNodes":[3,21]},{"coordinate":[7,2],"connectedNodes":[22,20]},{"coordinate":[6,2],"connectedNodes":[23,21]},{"coordinate":[5,2],"connectedNodes":[24,22]},{"coordinate":[4,2],"connectedNodes":[9,23]},{"coordinate":[0,1],"connectedNodes":[12]}]',
        origin='{"coordinate":[9,5],"direction":"W"}',
        destinations="[[0,1]]",
        default=True,
        blocklyEnabled=False,
        theme_name="grass",
        character_name="Van",
    )
    level_119 = Level.objects.create(
        name="119",
        episode=episode_12,
        next_level=level_120,
        path='[{"coordinate":[2,3],"connectedNodes":[1]},{"coordinate":[2,2],"connectedNodes":[0,2]},{"coordinate":[3,2],"connectedNodes":[1,3]},{"coordinate":[3,1],"connectedNodes":[2,4]},{"coordinate":[4,1],"connectedNodes":[3,5]},{"coordinate":[5,1],"connectedNodes":[4,6]},{"coordinate":[5,2],"connectedNodes":[7,5]},{"coordinate":[6,2],"connectedNodes":[6,8]},{"coordinate":[6,3],"connectedNodes":[9,7]},{"coordinate":[6,4],"connectedNodes":[10,8]},{"coordinate":[5,4],"connectedNodes":[11,9]},{"coordinate":[5,5],"connectedNodes":[12,10]},{"coordinate":[4,5],"connectedNodes":[13,11]},{"coordinate":[3,5],"connectedNodes":[12]}]',
        origin='{"coordinate":[2,3],"direction":"S"}',
        destinations="[[3,5]]",
        default=True,
        blocklyEnabled=False,
        theme_name="grass",
        character_name="Van",
    )
    level_118 = Level.objects.create(
        name="118",
        episode=episode_12,
        next_level=level_119,
        path='[{"coordinate":[1,3],"connectedNodes":[1]},{"coordinate":[2,3],"connectedNodes":[0,2]},{"coordinate":[3,3],"connectedNodes":[1,3]},{"coordinate":[3,4],"connectedNodes":[4,2]},{"coordinate":[4,4],"connectedNodes":[3,5]},{"coordinate":[4,5],"connectedNodes":[6,4]},{"coordinate":[5,5],"connectedNodes":[5,7]},{"coordinate":[5,6],"connectedNodes":[8,6]},{"coordinate":[6,6],"connectedNodes":[7,9]},{"coordinate":[7,6],"connectedNodes":[8,10]},{"coordinate":[8,6],"connectedNodes":[9]}]',
        origin='{"coordinate":[1,3],"direction":"E"}',
        destinations="[[8,6]]",
        default=True,
        pythonEnabled=False,
        pythonViewEnabled=True,
        theme_name="snow",
        character_name="Van",
        model_solution="[11]",
    )
    level_117 = Level.objects.create(
        name="117",
        episode=episode_12,
        next_level=level_118,
        path='[{"coordinate":[2,2],"connectedNodes":[1]},{"coordinate":[3,2],"connectedNodes":[0,2]},{"coordinate":[3,3],"connectedNodes":[3,1]},{"coordinate":[4,3],"connectedNodes":[2,4]},{"coordinate":[4,4],"connectedNodes":[5,3]},{"coordinate":[5,4],"connectedNodes":[4,6]},{"coordinate":[5,5],"connectedNodes":[7,5]},{"coordinate":[6,5],"connectedNodes":[6,8]},{"coordinate":[6,6],"connectedNodes":[9,7]},{"coordinate":[7,6],"connectedNodes":[8]}]',
        origin='{"coordinate":[2,2],"direction":"E"}',
        destinations="[[7,6]]",
        default=True,
        pythonEnabled=False,
        pythonViewEnabled=True,
        theme_name="snow",
        character_name="Van",
        model_solution="[8]",
    )
    level_116 = Level.objects.create(
        name="116",
        episode=episode_12,
        next_level=level_117,
        path='[{"coordinate":[1,4],"connectedNodes":[1]},{"coordinate":[2,4],"connectedNodes":[0,2]},{"coordinate":[3,4],"connectedNodes":[1,3]},{"coordinate":[4,4],"connectedNodes":[2,4]},{"coordinate":[5,4],"connectedNodes":[3,5]},{"coordinate":[6,4],"connectedNodes":[4]}]',
        origin='{"coordinate":[1,4],"direction":"E"}',
        destinations="[[6,4]]",
        default=True,
        pythonEnabled=False,
        theme_name="grass",
        character_name="Van",
        model_solution="[7]",
    )
    level_115 = Level.objects.create(
        name="115",
        episode=episode_12,
        next_level=level_116,
        path='[{"coordinate":[9,2],"connectedNodes":[1]},{"coordinate":[8,2],"connectedNodes":[26,0,2]},{"coordinate":[8,1],"connectedNodes":[3,1]},{"coordinate":[7,1],"connectedNodes":[4,2]},{"coordinate":[6,1],"connectedNodes":[5,20,3]},{"coordinate":[5,1],"connectedNodes":[6,4]},{"coordinate":[4,1],"connectedNodes":[7,5]},{"coordinate":[3,1],"connectedNodes":[8,6]},{"coordinate":[3,2],"connectedNodes":[9,7]},{"coordinate":[4,2],"connectedNodes":[8,10]},{"coordinate":[4,3],"connectedNodes":[11,9]},{"coordinate":[3,3],"connectedNodes":[12,10]},{"coordinate":[3,4],"connectedNodes":[13,11]},{"coordinate":[3,5],"connectedNodes":[14,12]},{"coordinate":[4,5],"connectedNodes":[13,15]},{"coordinate":[5,5],"connectedNodes":[14,16,17]},{"coordinate":[5,6],"connectedNodes":[21,15]},{"coordinate":[5,4],"connectedNodes":[15,18]},{"coordinate":[6,4],"connectedNodes":[17,19]},{"coordinate":[6,3],"connectedNodes":[18,27,20]},{"coordinate":[6,2],"connectedNodes":[19,4]},{"coordinate":[6,6],"connectedNodes":[16,22]},{"coordinate":[7,6],"connectedNodes":[21,23]},{"coordinate":[8,6],"connectedNodes":[22,24]},{"coordinate":[8,5],"connectedNodes":[23,25]},{"coordinate":[8,4],"connectedNodes":[24,26]},{"coordinate":[8,3],"connectedNodes":[27,25,1]},{"coordinate":[7,3],"connectedNodes":[19,26]}]',
        origin='{"coordinate":[9,2],"direction":"W"}',
        destinations="[[4,5]]",
        default=True,
        blocklyEnabled=False,
        theme_name="grass",
        character_name="Van",
    )
    level_114 = Level.objects.create(
        name="114",
        episode=episode_12,
        next_level=level_115,
        path='[{"coordinate":[0,2],"connectedNodes":[1]},{"coordinate":[1,2],"connectedNodes":[0,2]},{"coordinate":[1,3],"connectedNodes":[3,1]},{"coordinate":[1,4],"connectedNodes":[4,2]},{"coordinate":[1,5],"connectedNodes":[5,3]},{"coordinate":[2,5],"connectedNodes":[4,6]},{"coordinate":[2,4],"connectedNodes":[5,7]},{"coordinate":[3,4],"connectedNodes":[6,8]},{"coordinate":[3,3],"connectedNodes":[7,9]},{"coordinate":[3,2],"connectedNodes":[8,10]},{"coordinate":[3,1],"connectedNodes":[9,11]},{"coordinate":[4,1],"connectedNodes":[10,12]},{"coordinate":[4,2],"connectedNodes":[13,11]},{"coordinate":[5,2],"connectedNodes":[12]}]',
        origin='{"coordinate":[0,2],"direction":"E"}',
        destinations="[[5,2]]",
        default=True,
        blocklyEnabled=False,
        theme_name="grass",
        character_name="Van",
    )
    level_113 = Level.objects.create(
        name="113",
        episode=episode_12,
        next_level=level_114,
        path='[{"coordinate":[4,7],"connectedNodes":[1]},{"coordinate":[4,6],"connectedNodes":[0,2]},{"coordinate":[4,5],"connectedNodes":[3,1]},{"coordinate":[3,5],"connectedNodes":[2,4]},{"coordinate":[3,4],"connectedNodes":[3,5]},{"coordinate":[3,3],"connectedNodes":[4]}]',
        origin='{"coordinate":[4,7],"direction":"S"}',
        destinations="[[3,3]]",
        default=True,
        blocklyEnabled=False,
        theme_name="farm",
        character_name="Van",
    )
    level_112 = Level.objects.create(
        name="112",
        episode=episode_12,
        next_level=level_113,
        path='[{"coordinate":[2,7],"connectedNodes":[1]},{"coordinate":[2,6],"connectedNodes":[0,2]},{"coordinate":[3,6],"connectedNodes":[1,3,17]},{"coordinate":[4,6],"connectedNodes":[2,4]},{"coordinate":[5,6],"connectedNodes":[3,5]},{"coordinate":[6,6],"connectedNodes":[4,6]},{"coordinate":[6,5],"connectedNodes":[5,7]},{"coordinate":[6,4],"connectedNodes":[20,6,8]},{"coordinate":[6,3],"connectedNodes":[7,9]},{"coordinate":[6,2],"connectedNodes":[8,10]},{"coordinate":[6,1],"connectedNodes":[11,9]},{"coordinate":[5,1],"connectedNodes":[12,10]},{"coordinate":[4,1],"connectedNodes":[13,22,11]},{"coordinate":[3,1],"connectedNodes":[14,12]},{"coordinate":[2,1],"connectedNodes":[15,13]},{"coordinate":[1,1],"connectedNodes":[16,14]},{"coordinate":[0,1],"connectedNodes":[15]},{"coordinate":[3,5],"connectedNodes":[2,18]},{"coordinate":[3,4],"connectedNodes":[17,19]},{"coordinate":[4,4],"connectedNodes":[18,20,21]},{"coordinate":[5,4],"connectedNodes":[19,7]},{"coordinate":[4,3],"connectedNodes":[19,22]},{"coordinate":[4,2],"connectedNodes":[21,12]}]',
        origin='{"coordinate":[2,7],"direction":"S"}',
        destinations="[[0,1]]",
        default=True,
        pythonEnabled=False,
        pythonViewEnabled=True,
        theme_name="grass",
        character_name="Van",
        model_solution="[11]",
    )
    level_111 = Level.objects.create(
        name="111",
        episode=episode_12,
        next_level=level_112,
        path='[{"coordinate":[0,5],"connectedNodes":[1]},{"coordinate":[1,5],"connectedNodes":[0,2]},{"coordinate":[2,5],"connectedNodes":[1,3]},{"coordinate":[3,5],"connectedNodes":[2,4]},{"coordinate":[3,6],"connectedNodes":[5,3]},{"coordinate":[4,6],"connectedNodes":[4,6]},{"coordinate":[5,6],"connectedNodes":[5,7]},{"coordinate":[5,5],"connectedNodes":[6,8]},{"coordinate":[5,4],"connectedNodes":[7,9]},{"coordinate":[5,3],"connectedNodes":[8,10]},{"coordinate":[5,2],"connectedNodes":[9,11]},{"coordinate":[6,2],"connectedNodes":[10]}]',
        origin='{"coordinate":[0,5],"direction":"E"}',
        destinations="[[6,2]]",
        default=True,
        pythonEnabled=False,
        pythonViewEnabled=True,
        theme_name="grass",
        character_name="Van",
    )
    level_110 = Level.objects.create(
        name="110",
        episode=episode_12,
        next_level=level_111,
        path='[{"coordinate":[0,2],"connectedNodes":[2]},{"coordinate":[2,2],"connectedNodes":[2,3]},{"coordinate":[1,2],"connectedNodes":[0,1]},{"coordinate":[2,3],"connectedNodes":[4,1]},{"coordinate":[2,4],"connectedNodes":[5,3]},{"coordinate":[3,4],"connectedNodes":[4,6]},{"coordinate":[4,4],"connectedNodes":[5]}]',
        origin='{"coordinate":[0,2],"direction":"E"}',
        destinations="[[4,4]]",
        default=True,
        pythonEnabled=False,
        pythonViewEnabled=True,
        theme_name="farm",
        character_name="Van",
    )

    def set_decor(level, decor):
        set_decor_inner(level, decor, LevelDecor)

    set_decor(
        level_110,
        json.loads('[{"x":503,"y":409,"decorName":"tree2"},{"x":384,"y":503,"decorName":"bush"},{"x":355,"y":549,"decorName":"bush"},{"x":320,"y":503,"decorName":"bush"},{"x":399,"y":200,"decorName":"pond"},{"x":400,"y":299,"decorName":"pond"}]')
    )
    set_decor(
        level_111,
        json.loads('[{"x":611,"y":568,"decorName":"tree2"},{"x":595,"y":648,"decorName":"tree1"},{"x":714,"y":454,"decorName":"tree2"},{"x":652,"y":328,"decorName":"tree2"},{"x":600,"y":447,"decorName":"tree2"},{"x":651,"y":498,"decorName":"tree1"},{"x":366,"y":286,"decorName":"tree2"},{"x":285,"y":257,"decorName":"tree2"},{"x":206,"y":283,"decorName":"tree1"},{"x":82,"y":385,"decorName":"tree1"},{"x":46,"y":324,"decorName":"tree1"},{"x":233,"y":355,"decorName":"tree1"},{"x":637,"y":400,"decorName":"tree1"},{"x":87,"y":258,"decorName":"tree2"},{"x":291,"y":321,"decorName":"tree1"},{"x":164,"y":388,"decorName":"tree2"},{"x":129,"y":334,"decorName":"tree1"},{"x":128,"y":593,"decorName":"tree2"},{"x":235,"y":685,"decorName":"tree2"},{"x":103,"y":688,"decorName":"tree2"},{"x":181,"y":635,"decorName":"tree1"},{"x":70,"y":625,"decorName":"tree1"}]')
    )
    set_decor(
        level_112,
        json.loads('[{"x":50,"y":625,"decorName":"tree1"},{"x":50,"y":207,"decorName":"tree1"},{"x":50,"y":321,"decorName":"tree1"},{"x":50,"y":426,"decorName":"tree1"},{"x":50,"y":529,"decorName":"tree1"},{"x":398,"y":485,"decorName":"tree2"},{"x":500,"y":513,"decorName":"tree2"},{"x":500,"y":198,"decorName":"bush"},{"x":500,"y":274,"decorName":"bush"},{"x":500,"y":352,"decorName":"bush"}]')
    )
    set_decor(
        level_113,
        json.loads('[{"x":404,"y":415,"decorName":"tree2"},{"x":251,"y":420,"decorName":"bush"},{"x":187,"y":420,"decorName":"bush"},{"x":128,"y":420,"decorName":"bush"},{"x":221,"y":460,"decorName":"bush"},{"x":153,"y":460,"decorName":"bush"},{"x":183,"y":500,"decorName":"bush"},{"x":203,"y":596,"decorName":"pond"},{"x":203,"y":696,"decorName":"pond"}]')
    )
    set_decor(
        level_114,
        json.loads('[{"x":2,"y":438,"decorName":"tree1"},{"x":2,"y":543,"decorName":"tree1"},{"x":2,"y":329,"decorName":"tree1"},{"x":548,"y":556,"decorName":"tree1"},{"x":563,"y":469,"decorName":"tree1"},{"x":475,"y":549,"decorName":"tree2"},{"x":419,"y":514,"decorName":"tree1"},{"x":488,"y":434,"decorName":"tree2"}]')
    )
    set_decor(
        level_115,
        json.loads('[{"x":885,"y":343,"decorName":"tree1"},{"x":885,"y":440,"decorName":"tree1"},{"x":406,"y":383,"decorName":"tree2"},{"x":498,"y":306,"decorName":"tree2"},{"x":484,"y":190,"decorName":"tree2"}]')
    )
    set_decor(
        level_116,
        json.loads('[{"x":220,"y":305,"decorName":"pond"},{"x":539,"y":569,"decorName":"tree2"},{"x":569,"y":516,"decorName":"tree1"},{"x":498,"y":497,"decorName":"tree2"},{"x":182,"y":165,"decorName":"tree2"},{"x":225,"y":228,"decorName":"tree1"},{"x":159,"y":245,"decorName":"tree1"},{"x":312,"y":254,"decorName":"tree2"}]')
    )
    set_decor(
        level_117,
        json.loads('[{"x":443,"y":578,"decorName":"pond"},{"x":412,"y":648,"decorName":"tree1"},{"x":516,"y":651,"decorName":"tree2"},{"x":694,"y":341,"decorName":"tree1"},{"x":647,"y":292,"decorName":"tree2"},{"x":623,"y":362,"decorName":"tree1"},{"x":640,"y":434,"decorName":"tree2"}]')
    )
    set_decor(
        level_118,
        json.loads('[{"x":711,"y":507,"decorName":"tree2"},{"x":605,"y":508,"decorName":"tree2"},{"x":193,"y":396,"decorName":"bush"},{"x":250,"y":398,"decorName":"bush"}]')
    )
    set_decor(
        level_119,
        json.loads('[{"x":250,"y":391,"decorName":"tree2"},{"x":367,"y":237,"decorName":"tree1"},{"x":463,"y":317,"decorName":"tree2"},{"x":364,"y":395,"decorName":"tree1"}]')
    )
    set_decor(
        level_120,
        json.loads('[{"x":189,"y":337,"decorName":"pond"},{"x":793,"y":368,"decorName":"tree1"},{"x":814,"y":285,"decorName":"tree2"},{"x":869,"y":389,"decorName":"tree2"},{"x":334,"y":413,"decorName":"tree1"},{"x":403,"y":402,"decorName":"tree2"},{"x":192,"y":410,"decorName":"tree1"}]')
    )
    set_decor(
        level_121,
        json.loads('[{"x":383,"y":500,"decorName":"pond"},{"x":700,"y":500,"decorName":"pond"},{"x":700,"y":400,"decorName":"pond"},{"x":700,"y":300,"decorName":"pond"},{"x":550,"y":433,"decorName":"bush"},{"x":549,"y":467,"decorName":"bush"},{"x":550,"y":395,"decorName":"bush"},{"x":186,"y":431,"decorName":"tree1"},{"x":196,"y":387,"decorName":"tree1"},{"x":541,"y":501,"decorName":"tree2"}]')
    )
    set_decor(
        level_122,
        json.loads('[{"x":150,"y":320,"decorName":"bush"},{"x":350,"y":320,"decorName":"bush"},{"x":550,"y":320,"decorName":"bush"}]')
    )

    def set_blocks(level, blocks):
        set_blocks_inner(level, blocks, LevelBlock, Block)

    set_blocks(
        level_110,
        json.loads('[{"type":"move_forwards"},{"type":"turn_left"},{"type":"turn_right"}]'),
    )
    set_blocks(
        level_111,
        json.loads('[{"type":"move_forwards"},{"type":"turn_left"},{"type":"turn_right"}]'),
    )
    set_blocks(
        level_112,
        json.loads('[{"type":"move_forwards"},{"type":"turn_left"},{"type":"turn_right"}]'),
    )
    set_blocks(
        level_116,
        json.loads('[{"type":"move_forwards"},{"type":"controls_repeat_while"},{"type":"variables_numeric_set"},{"type":"variables_increment"},{"type":"variables_get"},{"type":"math_number"},{"type":"logic_compare"}]'),
    )
    set_blocks(
        level_117,
        json.loads('[{"type":"turn_left"},{"type":"turn_right"},{"type":"controls_repeat_while"},{"type":"variables_numeric_set"},{"type":"variables_increment"},{"type":"variables_get"},{"type":"math_number"},{"type":"logic_compare"}]'),
    )
    set_blocks(
        level_118,
        json.loads('[{"type":"move_forwards"},{"type":"turn_left"},{"type":"turn_right"},{"type":"controls_repeat_while"},{"type":"variables_numeric_set"},{"type":"variables_increment"},{"type":"variables_get"},{"type":"math_number"},{"type":"logic_compare"}]'),
    )

    level_109 = Level.objects.get(name="109")
    level_109.next_level = level_110
    level_109.save()


def delete_loop_levels(apps: Apps, *args):
    Level = apps.get_model("game", "Level")
    Episode = apps.get_model("game", "Episode")

    Level.objects.filter(episode__pk__in=range(12, 16)).delete()
    Episode.objects.filter(pk__in=range(12, 16)).delete()


class Migration(migrations.Migration):
    dependencies = [("game", "0085_add_new_blocks")]
    operations = [
        migrations.RunPython(
            add_loop_levels,
            reverse_code=delete_loop_levels,
        )
    ]



import json
from datetime import datetime
from unittest.mock import patch

from common.models import Teacher
from common.tests.utils.classes import create_class_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from deploy import captcha
from django.test.client import Client
from django.test.testcases import TestCase
from django.urls import reverse

from game.models import Level
from game.permissions import can_play_level
from game.tests.utils.level import create_save_level
from game.tests.utils.teacher import add_teacher_to_school, create_school
from game.views.level import _next_level_url, _prev_level_url


class LevelSelectionTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(LevelSelectionTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(LevelSelectionTestCase, cls).tearDownClass()

    def setUp(self):
        self.client = Client()

    def login(self, email, password):
        self.client.post(
            reverse("teacher_login"),
            {
                "auth-username": email,
                "auth-password": password,
                "teacher_login_view-current_step": "auth",
            },
            follow=True,
        )

    def logout(self):
        self.client.post(reverse("logout_view"), follow=True)

    def student_login(self, name, access_code, password):
        self.client.post(
            reverse("student_login", kwargs={"access_code": access_code}),
            {"username": name, "password": password},
            follow=True,
        )

    def level_data(self, levelID):
        data = {
            "origin": '{"coordinate":[3,5],"direction":"S"}',
            "pythonEnabled": False,
            "decor": [],
            "blocklyEnabled": True,
            "blocks": [
                {"type": "move_forwards"},
                {"type": "turn_left"},
                {"type": "turn_right"},
            ],
            "max_fuel": "50",
            "pythonViewEnabled": False,
            "character": "3",
            "name": f"level{levelID}",
            "theme": 1,
            "anonymous": False,
            "cows": "[]",
            "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],"connectedNodes":[0]}]',
            "traffic_lights": "[]",
            "destinations": "[[3,4]]",
        }
        return json.dumps(data)

    def test_list_episodes(self):
        url = reverse("levels")
        response = self.client.get(url)

        assert response.status_code == 200
        assert response.context["blocklyEpisodes"][0]["name"] == "Getting Started"
        assert response.context["blocklyEpisodes"][0]["levels"][0]["title"] == "Can you help the van get to the house?"

    def test_custom_levels_access(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()

        teacher1 = Teacher.objects.get(new_user__email=email1)
        teacher2 = Teacher.objects.get(new_user__email=email2)

        school1 = create_school()
        add_teacher_to_school(teacher1, school1, is_admin=True)
        add_teacher_to_school(teacher2, school1)

        # Create a class and a student for each teacher
        _, class_name1, access_code1 = create_class_directly(email1)
        _, class_name2, access_code2 = create_class_directly(email2)
        student_name1, student_password1, student1 = create_school_student_directly(access_code1)
        student_name2, student_password2, student2 = create_school_student_directly(access_code2)

        save_url = "save_level_for_editor"

        # Login as the second teacher
        self.login(email2, password2)
        teacher2_level = create_save_level(teacher2)
        save_level_url = reverse(save_url)

        response = self.client.post(save_level_url, {"data": self.level_data(teacher2_level.id)})

        assert response.status_code == 200

        # Login as the first student
        self.logout()
        self.student_login(student_name1, access_code1, student_password1)

        student1_level = create_save_level(student1)

        response = self.client.post(save_level_url, {"data": self.level_data(student1_level.id)})

        assert response.status_code == 200

        # Login as the second student
        self.logout()
        self.student_login(student_name2, access_code2, student_password2)

        student2_level = create_save_level(student2)

        response = self.client.post(save_level_url, {"data": self.level_data(student2_level.id)})

        assert response.status_code == 200

        # Login as first teacher again and check they have access to all the levels created above
        self.logout()
        self.login(email1, password1)

        url = reverse("levels")
        response = self.client.get(url)

        assert response.status_code == 200
        assert len(response.context["directly_shared_levels"]) == 1
        assert response.context["directly_shared_levels"][0]["owner"] == student1.new_user
        assert response.context["indirectly_shared_levels"][teacher2.new_user]
        assert len(response.context["indirectly_shared_levels"][teacher2.new_user]) == 2
        assert response.context["indirectly_shared_levels"][teacher2.new_user][0]["owner"] == teacher2.new_user
        assert response.context["indirectly_shared_levels"][teacher2.new_user][1]["owner"] == student2.new_user

        # Login as second teacher again and check they have access to only their student's level
        self.logout()
        self.login(email2, password2)

        url = reverse("levels")
        response = self.client.get(url)

        assert response.status_code == 200
        assert len(response.context["directly_shared_levels"]) == 1
        assert response.context["directly_shared_levels"][0]["owner"] == student2.new_user
        assert response.context["indirectly_shared_levels"] == {}

    def test_cannot_access_locked_level(self):
        email, password = signup_teacher_directly()

        teacher = Teacher.objects.get(new_user__email=email)

        school = create_school()
        add_teacher_to_school(teacher, school, is_admin=True)

        class1, _, access_code1 = create_class_directly(email)
        class2, _, access_code2 = create_class_directly(email)
        _, _, student1 = create_school_student_directly(access_code1)
        _, _, student2 = create_school_student_directly(access_code2)

        level1 = Level.objects.get(id=1)

        level1.locked_for_class.add(class1)

        assert not can_play_level(student1.new_user, level1, False)
        assert can_play_level(student2.new_user, level1, False)

    def test_next_level_for_locked_levels(self):
        email, password = signup_teacher_directly()

        teacher = Teacher.objects.get(new_user__email=email)

        school = create_school()
        add_teacher_to_school(teacher, school, is_admin=True)

        klass, _, access_code = create_class_directly(email)
        _, _, student = create_school_student_directly(access_code)

        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")
        level3 = Level.objects.get(name="3")
        level4 = Level.objects.get(name="4")
        level106 = Level.objects.get(name="106")
        level107 = Level.objects.get(name="107")
        level108 = Level.objects.get(name="108")
        level109 = Level.objects.get(name="109")

        level2.locked_for_class.add(klass)
        level3.locked_for_class.add(klass)
        level107.locked_for_class.add(klass)
        level108.locked_for_class.add(klass)
        level109.locked_for_class.add(klass)

        next_level_url = _next_level_url(level1, student.new_user, False)

        assert next_level_url == f"/rapidrouter/{level4.name}/"

        prev_level_url = _prev_level_url(level4, student.new_user, False)
        assert prev_level_url == f"/rapidrouter/{level1.name}/"

        next_level_url = _next_level_url(level106, student.new_user, False)

        assert next_level_url == f"/rapidrouter/{level109.name}/"

        prev_level_url = _prev_level_url(level109, student.new_user, False)
        assert prev_level_url == f"/rapidrouter/{level106.name}/"

    @patch("game.views.level.datetime", side_effect=lambda *args, **kw: datetime(*args, **kw))
    def test_xmas_theme(self, mock_datetime):
        november = datetime(2023, 11, 1, 0, 0, 0, 0)
        mock_datetime.now.return_value = november

        response = self.client.get(f"{reverse('home')}/rapidrouter/1/")
        assert """CHARACTER_NAME = "Van"\n""" in response.content.decode("utf-8")
        assert """CHARACTER_URL = "characters/top_view/Van.svg"\n""" in response.content.decode("utf-8")
        assert """WRECKAGE_URL = "van_wreckage.svg"\n""" in response.content.decode("utf-8")
        assert """BACKGROUND_URL = "decor/grass/tile1.svg"\n""" in response.content.decode("utf-8")
        assert """HOUSE_URL = "decor/grass/house.svg"\n""" in response.content.decode("utf-8")
        assert """CFC_URL = "decor/grass/cfc.svg"\n""" in response.content.decode("utf-8")

        december = datetime(2023, 12, 1, 0, 0, 0, 0)
        mock_datetime.now.return_value = december

        response = self.client.get(f"{reverse('home')}/rapidrouter/1/")
        assert """CHARACTER_NAME = "Van"\n""" in response.content.decode("utf-8")
        assert """CHARACTER_URL = "characters/top_view/Sleigh.svg"\n""" in response.content.decode("utf-8")
        assert """WRECKAGE_URL = "sleigh_wreckage.svg"\n""" in response.content.decode("utf-8")
        assert """BACKGROUND_URL = "decor/snow/tile1.svg"\n""" in response.content.decode("utf-8")
        assert """HOUSE_URL = "decor/snow/house.svg"\n""" in response.content.decode("utf-8")
        assert """CFC_URL = "decor/snow/cfc.svg"\n""" in response.content.decode("utf-8")



from django.db import migrations


def add_new_blocks(apps, schema_editor):
    Block = apps.get_model("game", "Block")

    VARIABLE = 5
    MATH = 6

    block1 = Block(type="variables_set", block_type=VARIABLE)
    block2 = Block(type="variables_numeric_set", block_type=VARIABLE)
    block3 = Block(type="variables_increment", block_type=VARIABLE)
    block4 = Block(type="variables_get", block_type=VARIABLE)
    block5 = Block(type="math_number", block_type=MATH)
    block6 = Block(type="math_arithmetic", block_type=MATH)
    block7 = Block(type="logic_compare", block_type=MATH)

    block1.save()
    block2.save()
    block3.save()
    block4.save()
    block5.save()
    block6.save()
    block7.save()


def remove_new_blocks(apps, schema_editor):
    Block = apps.get_model("game", "Block")

    block1 = Block.objects.get(type="variables_set")
    block2 = Block.objects.get(type="variables_numeric_set")
    block3 = Block.objects.get(type="variables_increment")
    block4 = Block.objects.get(type="variables_get")
    block5 = Block.objects.get(type="math_number")
    block6 = Block.objects.get(type="math_arithmetic")
    block7 = Block.objects.get(type="logic_compare")

    block1.delete()
    block2.delete()
    block3.delete()
    block4.delete()
    block5.delete()
    block6.delete()
    block7.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("game", "0084_alter_block_block_type"),
    ]

    operations = [migrations.RunPython(add_new_blocks, reverse_code=remove_new_blocks)]



import json
import random
import itertools
from django.test import TestCase

from game.random_road import generate_random_map_data


def find_node(x, y, nodes):
    for node in nodes:
        if node["coordinate"][0] == x and node["coordinate"][1] == y:
            return True
    return False


def check_direction(node, neighbour):
    if neighbour["coordinate"][1] < node["coordinate"][1]:
        direction = "N"
    elif neighbour["coordinate"][0] < node["coordinate"][0]:
        direction = "E"
    elif neighbour["coordinate"][1] > node["coordinate"][1]:
        direction = "S"
    elif neighbour["coordinate"][0] > node["coordinate"][0]:
        direction = "W"
    return direction


class RandomRoadTestCase(TestCase):
    def create_test_data(
        self,
        num_tiles=None,
        branchiness=None,
        loopiness=None,
        curviness=None,
        traffic_lights_enabled=False,
        decor_enabled=True,
        cows_enabled=False,
    ):
        if num_tiles is None:
            num_tiles = random.randint(3, 40)
        if branchiness is None:
            branchiness = random.random()
        if loopiness is None:
            loopiness = random.random()
        if curviness is None:
            curviness = random.random()

        return generate_random_map_data(
            num_tiles,
            branchiness,
            loopiness,
            curviness,
            traffic_lights_enabled,
            decor_enabled,
            cows_enabled,
        )

    def check_if_valid(self, origin, destinations, path, decor):
        check = find_node(origin["coordinate"][0], origin["coordinate"][1], path)
        self.assertTrue(check)

        for node in destinations:
            self.assertTrue(find_node(node[0], node[1], path))

        if decor:
            # Check if any decor got generated
            self.assertTrue(decor)

    def test_tiles_less_or_equal_size(self):
        """Test that the number of tiles is less or equal to the size parameter"""

        number_of_tiles = 30
        data = self.create_test_data(num_tiles=number_of_tiles)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if number of tiles = size
        self.assertTrue(len(path) <= number_of_tiles)

    def test_branchiness_min(self):
        """Test that if the branchiness is 0 we don't get branches."""

        data = self.create_test_data(branchiness=0, loopiness=0)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if no branches
        for node in path:
            self.assertTrue(len(node["connectedNodes"]) <= 2)

    def test_branchiness_max(self):
        """Test that if the branchiness > 0 we get branches."""

        data = self.create_test_data(branchiness=2, loopiness=0)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if some branches
        branch_count = 0
        for node in path:
            if len(node["connectedNodes"]) > 2:
                branch_count += 1

        self.assertTrue(branch_count)

    def test_curviness_max(self):
        """Test that the path is curving significantly if the curviness is set to max value."""

        data = self.create_test_data(curviness=1.0)
        path = json.loads(data["path"])

        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        num_curves = 0
        for node in path:
            already_checked_for_curves = (
                []
            )  # So that A->B->C and C->B->A don't count as two separate curves
            if len(node["connectedNodes"]) > 1:
                for (one_way, other_way) in itertools.combinations(
                    node["connectedNodes"], 2
                ):
                    if (
                        one_way,
                        other_way,
                    ) not in already_checked_for_curves:
                        already_checked_for_curves.append(
                            (
                                one_way,
                                other_way,
                            )
                        )
                        already_checked_for_curves.append(
                            (
                                other_way,
                                one_way,
                            )
                        )

                        previous_to_this = check_direction(path[one_way], node)
                        this_to_next = check_direction(node, path[other_way])

                        num_curves += int(not (previous_to_this == this_to_next))

        if len(path) == 3:
            self.assertEqual(num_curves, 1)
        else:
            self.assertGreaterEqual(num_curves, 2)

    def test_traffic_light_no(self):
        """Test that traffic lights don't get generated"""

        data = self.create_test_data(traffic_lights_enabled=False)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        traffic_lights = json.loads(data["traffic_lights"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no traffic lights got generated.
        self.assertFalse(traffic_lights)

    def test_traffic_light_yes(self):
        """Test that traffic lights get generated"""

        data = self.create_test_data(traffic_lights_enabled=True)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        traffic_lights = json.loads(data["traffic_lights"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no traffic lights got generated.
        self.assertTrue(traffic_lights)

    def test_decor_no(self):
        """Test that scenery doesn't get generated"""

        data = self.create_test_data(decor_enabled=False)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no scenery got generated.
        self.assertFalse(decor)

    def test_decor_yes(self):
        """Test that scenery does get generated"""

        data = self.create_test_data(decor_enabled=True)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert scenery got generated.
        self.assertTrue(decor)



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ("game", "0029_level_pythonviewenabled"),
        ("game", "0029_disable_route_scores"),
    ]

    operations = []



from django.urls import reverse
from hamcrest import *
from hamcrest.core.base_matcher import BaseMatcher
from rest_framework import status
from rest_framework.test import APITestCase

from game.character import get_all_character
from game.decor import get_all_decor
from game.theme import get_all_themes
from .utils.user import get_superuser


class APITests(APITestCase):
    def test_list_decors(self):
        url = reverse("decor-list")
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_decor())))

    def test_known_decor_detail(self):
        decor_id = 1
        url = reverse("decor-detail", kwargs={"pk": decor_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(decor_id))

    def test_unknown_decor_detail(self):
        decor_id = 0
        url = reverse("decor-detail", kwargs={"pk": decor_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_levels_for_known_episode(self):
        episode_id = 1
        url = reverse("level-for-episode", kwargs={"pk": episode_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(greater_than(0)))

    def test_levels_for_unknown_episode(self):
        episode_id = 0
        url = reverse("level-for-episode", kwargs={"pk": episode_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(0))

    def test_list_themes(self):
        url = reverse("theme-list")
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_themes())))

    def test_known_theme_detail(self):
        theme_id = 1
        url = reverse("theme-detail", kwargs={"pk": theme_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(theme_id))

    def test_unknown_theme_detail(self):
        theme_id = 0
        url = reverse("theme-detail", kwargs={"pk": theme_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_list_characters(self):
        url = reverse("character-list")
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_character())))

    def test_known_character_detail(self):
        character_id = 1
        url = reverse("character-detail", kwargs={"pk": character_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(character_id))

    def test_unknown_character_detail(self):
        character_id = 0
        url = reverse("character-detail", kwargs={"pk": character_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_list_episodes(self):
        url = reverse("episode-list")
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data[0]["name"], equal_to("Getting Started"))

    def test_episode_details(self):
        episode_id = 1
        url = reverse("episode-detail", kwargs={"pk": episode_id})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["name"], equal_to("Getting Started"))


def has_status_code(status_code):
    return HasStatusCode(status_code)


class HasStatusCode(BaseMatcher):
    def __init__(self, status_code):
        self.status_code = status_code

    def _matches(self, response):
        return response.status_code == self.status_code

    def describe_to(self, description):
        description.append_text("has status code ").append_text(self.status_code)

    def describe_mismatch(self, response, mismatch_description):
        mismatch_description.append_text("had status code ").append_text(
            response.status_code
        )



from aimmo import urls as aimmo_urls
from django.conf.urls import include, url
from django.contrib import admin
from django.urls import path
from portal import urls as portal_urls

from game import urls as game_urls

admin.autodiscover()

urlpatterns = [
    url(r"^", include(portal_urls)),
    path("administration/", admin.site.urls),
    url(r"^rapidrouter/", include(game_urls)),
    url(r"^aimmo/", include(aimmo_urls)),
]



from __future__ import print_function

import os
import time
from builtins import str

from django.urls import reverse
from hamcrest import assert_that, equal_to, contains_string, ends_with
from portal.tests.pageObjects.portal.base_page import BasePage
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.expected_conditions import (
    presence_of_all_elements_located,
)
from selenium.webdriver.support.ui import WebDriverWait


class GamePage(BasePage):
    def __init__(self, browser):
        super(GamePage, self).__init__(browser)

        assert self.on_correct_page("game_page")

        self._dismiss_initial_dialog()

    def _dismiss_initial_dialog(self):
        self.dismiss_dialog("play_button")
        return self

    def dismiss_dialog(self, button_id):
        self.wait_for_element_to_be_clickable((By.ID, button_id), wait_seconds=15)
        self.browser.find_element(By.ID, button_id).click()
        self.wait_for_element_to_be_invisible((By.ID, button_id), wait_seconds=15)

    def load_solution(self, workspace_id):
        self.browser.find_element(By.ID, "load_tab").click()
        selector = "#loadWorkspaceTable tr[value='" + str(workspace_id) + "']"
        self.wait_for_element_to_be_clickable((By.CSS_SELECTOR, selector))
        self.browser.find_element(By.CSS_SELECTOR, selector).click()
        self.browser.find_element(By.ID, "loadWorkspace").click()
        time.sleep(1)
        return self

    def clear(self):
        self.browser.find_element(By.ID, "clear_tab").click()
        return self

    def try_again(self):
        self.dismiss_dialog("try_again_button")
        return self

    def step(self):
        self.browser.find_element(By.ID, "step_tab").click()
        return self

    def solution_button(self):
        self.browser.find_element(By.ID, "solution_tab").click()
        solution_loaded = self.browser.execute_script("return ocargo.solutionLoaded;")
        timeout = time.time() + 30

        while not solution_loaded:
            solution_loaded = self.browser.execute_script(
                "return ocargo.solutionLoaded;"
            )
            if time.time() > timeout:
                break
        return self

    def python_commands_button(self):
        self.browser.find_element(By.ID, "van_commands_help").click()
        return self

    def clear_console_button(self):
        self.browser.find_element(By.ID, "clear_console").click()
        return self

    def assert_level_number(self, level_number):
        path = reverse("play_default_level", kwargs={"levelName": str(level_number)})
        assert_that(self.browser.current_url.replace("#myModal", ""), ends_with(path))

    def assert_episode_number(self, episode_number):
        path = reverse("start_episode", kwargs={"episodeId": str(episode_number)})
        assert_that(self.browser.current_url.replace("#myModal", ""), ends_with(path))

    def assert_is_green_light(self, traffic_light_index):
        self._assert_light_is_on(traffic_light_index, "green")

    def assert_is_red_light(self, traffic_light_index):
        self._assert_light_is_on(traffic_light_index, "red")

    def _assert_light_is_on(self, traffic_light_index, colour):
        image = self.browser.find_element(By.ID,
            "trafficLight_%s_%s" % (traffic_light_index, colour)
        )

        assert_that(image.get_attribute("opacity"), equal_to("1"))

    def run_program(self, wait_for_element_id="modal-content"):
        self.browser.find_element(By.ID, "fast_tab").click()

        try:
            self.wait_for_element_to_be_clickable((By.ID, wait_for_element_id), 45)
        except TimeoutException as e:
            millis = int(round(time.time() * 1000))
            screenshot_filename = "/tmp/game_tests_%s-%s.png" % (
                os.getenv("BUILD_NUMBER", "nonumber"),
                str(millis),
            )
            print("Saved screenshot to " + screenshot_filename)
            self.browser.get_screenshot_as_file(screenshot_filename)
            raise e

        return self

    def run_retry_program(self):
        self.run_program("try_again_button")
        modal_content = self.browser.find_element(By.ID, "modal-content").text
        assert_that(modal_content, contains_string("Try creating a simpler program."))
        self.browser.find_element(By.ID, "try_again_button").click()
        time.sleep(1)
        return self

    def run_crashing_program(self):
        return self._run_failing_program("What went wrong")

    def run_cow_crashing_program(self):
        return self._run_failing_program("You ran into a cow!")

    def run_out_of_instructions_program(self):
        return self._run_failing_program(
            "The van ran out of instructions before it reached a destination."
        )

    def run_out_of_fuel_program(self):
        return self._run_failing_program("You ran out of fuel!")

    def run_a_red_light_program(self):
        return self._run_failing_program(
            "Uh oh, you just sent the van through a red light!"
        )

    def run_not_delivered_everywhere_program(self):
        return self._run_failing_program(
            "There are destinations that have not been delivered to."
        )

    def run_undefined_procedure_program(self):
        return self._run_procedure_error_program(
            "Your program doesn't look quite right..."
        )

    def run_parse_error_program(self):
        return self._run_failing_python_program_console(
            "my_van.", "ParseError: bad input on line"
        )

    def run_attribute_error_program(self):
        return self._run_failing_python_program_console(
            "my_van.go()", "AttributeError: 'Van' object has no attribute"
        )

    def run_print_program(self):
        return self._run_failing_python_program_console(
            'print("hello world")', "hello world"
        )

    def run_invalid_import_program(self):
        return self._run_failing_python_program_popup(
            "from van import Va", "You can only import 'Van' from 'van'"
        )

    def check_python_commands(self):
        self.python_commands_button()
        time.sleep(1)
        python_commands = self.browser.find_element(By.ID, "myModal-lead").text
        assert_that(
            python_commands,
            contains_string("Run the following commands on the van object v"),
        )
        return self

    def write_to_then_clear_console(self):
        self._write_code("my_van.")
        self.browser.find_element(By.ID, "fast_tab").click()
        time.sleep(1)
        console = self.browser.find_element(By.ID, "consoleOutput")
        self.clear_console_button()
        assert_that(console.text == "")
        return self

    def next_episode(self):
        self.assert_success()
        self.browser.find_element(By.ID, "next_episode_button").click()
        WebDriverWait(self.browser, 10).until(
            presence_of_all_elements_located((By.ID, "blockly_tab"))
        )
        return self

    def next_level(self):
        self.assert_success()
        self.browser.find_element(By.ID, "next_level_button").click()
        WebDriverWait(self.browser, 10).until(
            presence_of_all_elements_located((By.ID, "blockly_tab"))
        )
        return self

    def _run_failing_program(self, text):
        self.run_program("try_again_button")
        error_message = self.browser.find_element(By.ID, "myModal-lead").text
        assert_that(error_message, contains_string(text))
        return self

    def _run_failing_python_program_console(self, code, console_message):
        self._write_code(code)
        self.browser.find_element(By.ID, "fast_tab").click()
        time.sleep(1)
        console = self.browser.find_element(By.ID, "consoleOutput")
        assert_that(console.text, contains_string(console_message))
        return self

    def _run_failing_python_program_popup(self, code, text):
        self._write_code(code)
        self.browser.find_element(By.ID, "fast_tab").click()
        time.sleep(1)
        error_message = self.browser.find_element(By.ID, "myModal-lead").text
        assert_that(error_message, contains_string(text))
        return self

    def _write_code(self, code):
        self.browser.execute_script(
            "ocargo.pythonControl.appendCode(arguments[0])", code
        )
        return self

    def _run_procedure_error_program(self, text):
        self.run_program("close_button")
        error_message = self.browser.find_element(By.ID, "myModal-mainText").text
        assert_that(error_message, contains_string(text))

    def _assert_score(self, element_id, score):
        score_text = self.browser.find_element(By.ID, element_id).text
        score_number = score_text.split("/")[0]
        assert_that(score_number, equal_to(str(score)))
        return self

    def assert_success(self):
        modal_content = self.browser.find_element(By.ID, "modal-content").text
        assert_that(modal_content, contains_string("Congratulations"))
        return self

    def assert_route_score(self, score):
        return self._assert_score("routeScore", score)

    def assert_algorithm_score(self, score):
        return self._assert_score("algorithmScore", score)



"""
    Decor data
"""

from builtins import object

from rest_framework.reverse import reverse

from game.theme import get_theme, get_all_themes


class Decor(object):
    def __init__(self, pk, name, url, xmas_url, width, height, theme, z_index):
        self.id = self.pk = pk
        self.name = name
        self.url = url
        self.xmas_url = xmas_url
        self.width = width
        self.height = height
        self.theme = theme
        self.z_index = z_index


DECOR_DATA = {
    ("tree1", "grass"): Decor(
        z_index=4,
        name="tree1",
        url="decor/grass/tree1.svg",
        xmas_url="decor/snow/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=1,
    ),
    ("tree2", "grass"): Decor(
        z_index=4,
        name="tree2",
        url="decor/grass/tree2.svg",
        xmas_url="decor/snow/tree2.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=2,
    ),
    ("bush", "grass"): Decor(
        z_index=3,
        name="bush",
        url="decor/grass/bush.svg",
        xmas_url="decor/snow/bush.svg",
        height=50,
        width=50,
        theme=get_theme("grass"),
        pk=3,
    ),
    ("house", "grass"): Decor(
        z_index=1,
        name="house",
        url="decor/grass/house.svg",
        xmas_url="decor/snow/house.svg",
        height=50,
        width=50,
        theme=get_theme("grass"),
        pk=4,
    ),
    ("cfc", "grass"): Decor(
        z_index=1,
        name="cfc",
        # FUTURE: add external branding option
        url="decor/grass/cfc.svg",
        xmas_url="decor/snow/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("grass"),
        pk=5,
    ),
    ("pond", "grass"): Decor(
        z_index=2,
        name="pond",
        url="decor/grass/pond.svg",
        xmas_url="decor/snow/pond.svg",
        height=100,
        width=150,
        theme=get_theme("grass"),
        pk=6,
    ),
    ("tree1", "snow"): Decor(
        z_index=4,
        name="tree1",
        url="decor/snow/tree1.svg",
        xmas_url="decor/snow/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=7,
    ),
    ("tree2", "snow"): Decor(
        z_index=4,
        name="tree2",
        url="decor/snow/tree2.svg",
        xmas_url="decor/snow/tree2.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=8,
    ),
    ("bush", "snow"): Decor(
        z_index=3,
        name="bush",
        url="decor/snow/bush.svg",
        xmas_url="decor/snow/bush.svg",
        height=50,
        width=50,
        theme=get_theme("snow"),
        pk=9,
    ),
    ("house", "snow"): Decor(
        z_index=1,
        name="house",
        url="decor/snow/house.svg",
        xmas_url="decor/snow/house.svg",
        height=50,
        width=50,
        theme=get_theme("snow"),
        pk=10,
    ),
    ("cfc", "snow"): Decor(
        z_index=1,
        name="cfc",
        # FUTURE: add external branding option
        url="decor/snow/cfc.svg",
        xmas_url="decor/snow/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("snow"),
        pk=11,
    ),
    ("pond", "snow"): Decor(
        z_index=2,
        name="pond",
        url="decor/snow/pond.svg",
        xmas_url="decor/snow/pond.svg",
        height=100,
        width=150,
        theme=get_theme("snow"),
        pk=12,
    ),
    ("tile1", "grass"): Decor(
        z_index=0,
        name="tile1",
        url="decor/grass/tile1.svg",
        xmas_url="decor/snow/tile1.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=13,
    ),
    ("tile1", "snow"): Decor(
        z_index=0,
        name="tile1",
        url="decor/snow/tile1.svg",
        xmas_url="decor/snow/tile1.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=14,
    ),
    ("tile2", "snow"): Decor(
        z_index=0,
        name="tile2",
        url="decor/snow/tile2.svg",
        xmas_url="decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=15,
    ),
    ("house", "farm"): Decor(
        z_index=1,
        name="house",
        url="decor/farm/house1.svg",
        xmas_url="decor/snow/house1.svg",
        height=224,
        width=184,
        theme=get_theme("farm"),
        pk=16,
    ),
    ("cfc", "farm"): Decor(
        z_index=1,
        name="cfc",
        url="decor/farm/cfc.svg",
        xmas_url="decor/snow/barn.svg",
        height=301,
        width=332,
        theme=get_theme("farm"),
        pk=17,
    ),
    ("bush", "farm"): Decor(
        z_index=3,
        name="bush",
        url="decor/farm/bush.svg",
        xmas_url="decor/snow/bush.svg",
        height=30,
        width=50,
        theme=get_theme("farm"),
        pk=18,
    ),
    ("tree1", "farm"): Decor(
        z_index=4,
        name="tree1",
        url="decor/farm/tree1.svg",
        xmas_url="decor/snow/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("farm"),
        pk=19,
    ),
    ("tree2", "farm"): Decor(
        z_index=4,
        name="tree2",
        url="decor/farm/house2.svg",
        xmas_url="decor/snow/house.svg",
        height=88,
        width=65,
        theme=get_theme("farm"),
        pk=20,
    ),
    ("pond", "farm"): Decor(
        z_index=2,
        name="pond",
        url="decor/farm/crops.svg",
        xmas_url="decor/snow/crops.svg",
        height=100,
        width=150,
        theme=get_theme("farm"),
        pk=21,
    ),
    ("tile1", "farm"): Decor(
        z_index=0,
        name="tile1",
        url="decor/farm/tile1.svg",
        xmas_url="decor/snow/tile1.svg",
        height=337,
        width=194,
        theme=get_theme("farm"),
        pk=22,
    ),
    ("tile1", "city"): Decor(
        z_index=0,
        name="tile1",
        url="decor/city/pavementTile.png",
        xmas_url="decor/snow/tile1.svg",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=23,
    ),
    ("house", "city"): Decor(
        z_index=1,
        name="house",
        url="decor/city/house.svg",
        xmas_url="decor/snow/house2.svg",
        height=50,
        width=50,
        theme=get_theme("city"),
        pk=24,
    ),
    ("bush", "city"): Decor(
        z_index=3,
        name="bush",
        url="decor/city/bush.svg",
        xmas_url="decor/snow/bush.svg",
        height=50,
        width=50,
        theme=get_theme("city"),
        pk=25,
    ),
    ("tree1", "city"): Decor(
        z_index=4,
        name="tree1",
        url="decor/city/shop.svg",
        xmas_url="decor/snow/shop.svg",
        height=70,
        width=70,
        theme=get_theme("city"),
        pk=26,
    ),
    ("tree2", "city"): Decor(
        z_index=4,
        name="tree2",
        url="decor/city/school.svg",
        xmas_url="decor/snow/school.svg",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=27,
    ),
    ("pond", "city"): Decor(
        z_index=2,
        name="pond",
        url="decor/city/hospital.svg",
        xmas_url="decor/snow/hospital.svg",
        height=157,
        width=140,
        theme=get_theme("city"),
        pk=28,
    ),
    ("cfc", "city"): Decor(
        z_index=1,
        name="cfc",
        # FUTURE: add external branding option
        url="decor/grass/cfc.svg",
        xmas_url="decor/snow/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("city"),
        pk=29,
    ),
    ("tile2", "grass"): Decor(
        z_index=0,
        name="tile2",
        url="decor/snow/tile2.svg",
        xmas_url="decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=30,
    ),
    ("tile2", "farm"): Decor(
        z_index=0,
        name="tile2",
        url="decor/snow/tile2.svg",
        xmas_url="decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("farm"),
        pk=31,
    ),
    ("tile2", "city"): Decor(
        z_index=0,
        name="tile2",
        url="decor/snow/tile2.svg",
        xmas_url="decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=32,
    ),
}


def get_decor_element(name, theme):
    """Helper method to get a decor element corresponding to the theme or a default one."""
    try:
        return DECOR_DATA[(name, theme.name)]
    except KeyError:
        for theme_object in get_all_themes():
            try:
                return DECOR_DATA[(name, theme_object.name)]
            except KeyError:
                pass
    raise KeyError


def get_all_decor():
    return list(DECOR_DATA.values())


def get_decor_element_by_pk(pk):
    for decor in list(DECOR_DATA.values()):
        if decor.pk == int(pk):
            return decor
    raise KeyError


def get_decors_url(pk, request):
    return reverse("decor-detail", args={pk}, request=request)



from django.contrib.auth.models import User


def get_superuser():
    """Get a superuser for testing, or create one if there isn't one."""
    try:
        return User.objects.get(username="superuser")
    except User.DoesNotExist:
        return User.objects.create_superuser(
            "superuser", "superuser@codeforlife.education", "password"
        )



from django.db import migrations


def update_level(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    level27 = Level.objects.get(name="27", default=1)
    level27.model_solution = "[13]"
    level27.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0066_rm_character_model")]

    operations = [migrations.RunPython(update_level)]
